# -*- coding: utf-8 -*-
"""
cyipopt: Python wrapper for the Ipopt optimization package, written in Cython.

Copyright (C) 2012-2015 Amit Aides
Copyright (C) 2015-2017 Matthias Kümmerer
Copyright (C) 2017-2021 cyipopt developers

License: EPL 1.0
"""

cdef extern from "IpStdCInterface.h":

    ctypedef double Number

    ctypedef int Index

    ctypedef int Int

    ctypedef struct IpoptProblemInfo:
        pass

    ctypedef IpoptProblemInfo* IpoptProblem

    ctypedef int Bool

    ctypedef void* UserDataPtr

    cdef enum ApplicationReturnStatus:
        Solve_Succeeded=0
        Solved_To_Acceptable_Level=1
        Infeasible_Problem_Detected=2
        Search_Direction_Becomes_Too_Small=3
        Diverging_Iterates=4
        User_Requested_Stop=5
        Feasible_Point_Found=6
        Maximum_Iterations_Exceeded=-1
        Restoration_Failed=-2
        Error_In_Step_Computation=-3
        Maximum_CpuTime_Exceeded=-4
        Not_Enough_Degrees_Of_Freedom=-10
        Invalid_Problem_Definition=-11
        Invalid_Option=-12
        Invalid_Number_Detected=-13
        Unrecoverable_Exception=-100
        NonIpopt_Exception_Thrown=-101
        Insufficient_Memory=-102
        Internal_Error=-199

    cdef enum AlgorithmMode:
        RegularMode=0
        RestorationPhaseMode=1

    ctypedef Bool (*Eval_F_CB)(
                    Index n,
                    Number* x,
                    Bool new_x,
                    Number* obj_value,
                    UserDataPtr user_data
                    )

    ctypedef Bool (*Eval_Grad_F_CB)(
                    Index n,
                    Number* x,
                    Bool new_x,
                    Number* grad_f,
                    UserDataPtr user_data
                    )

    ctypedef Bool (*Eval_G_CB)(
                    Index n,
                    Number* x,
                    Bool new_x,
                    Index m,
                    Number* g,
                    UserDataPtr user_data
                    )

    ctypedef Bool (*Eval_Jac_G_CB)(
                    Index n,
                    Number *x,
                    Bool new_x,
                    Index m,
                    Index nele_jac,
                    Index *iRow,
                    Index *jCol,
                    Number *values,
                    UserDataPtr user_data
                    )

    ctypedef Bool (*Eval_H_CB)(
                    Index n,
                    Number *x,
                    Bool new_x,
                    Number obj_factor,
                    Index m,
                    Number *lambd,
                    Bool new_lambda,
                    Index nele_hess,
                    Index *iRow,
                    Index *jCol,
                    Number *values,
                    UserDataPtr user_data
                    )

    ctypedef Bool (*Intermediate_CB)(
                    Index alg_mod,
                    Index iter_count,
                    Number obj_value,
                    Number inf_pr,
                    Number inf_du,
                    Number mu,
                    Number d_norm,
                    Number regularization_size,
                    Number alpha_du,
                    Number alpha_pr,
                    Index ls_trials,
                    UserDataPtr user_data
                    )

    IpoptProblem CreateIpoptProblem(
                            Index n,
                            Number* x_L,
                            Number* x_U,
                            Index m,
                            Number* g_L,
                            Number* g_U,
                            Index nele_jac,
                            Index nele_hess,
                            Index index_style,
                            Eval_F_CB eval_f,
                            Eval_G_CB eval_g,
                            Eval_Grad_F_CB eval_grad_f,
                            Eval_Jac_G_CB eval_jac_g,
                            Eval_H_CB eval_h
                            )

    void FreeIpoptProblem(IpoptProblem ipopt_problem)

    Bool AddIpoptStrOption(IpoptProblem ipopt_problem, char* keyword, char* val)

    Bool AddIpoptNumOption(IpoptProblem ipopt_problem, char* keyword, Number val)

    Bool AddIpoptIntOption(IpoptProblem ipopt_problem, char* keyword, Int val)

    Bool OpenIpoptOutputFile(IpoptProblem ipopt_problem, char* file_name, Int print_level)

    Bool SetIpoptProblemScaling(
                    IpoptProblem ipopt_problem,
                    Number obj_scaling,
                    Number* x_scaling,
                    Number* g_scaling
                    )

    Bool SetIntermediateCallback(
                    IpoptProblem ipopt_problem,
                    Intermediate_CB intermediate_cb
                    )

    ApplicationReturnStatus IpoptSolve(
                    IpoptProblem ipopt_problem,
                    Number* x,
                    Number* g,
                    Number* obj_val,
                    Number* mult_g,
                    Number* mult_x_L,
                    Number* mult_x_U,
                    UserDataPtr user_data
                    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
cyipopt: Python wrapper for the Ipopt optimization package, written in Cython.

Copyright (C) 2012-2015 Amit Aides
Copyright (C) 2015-2017 Matthias Kümmerer
Copyright (C) 2017-2021 cyipopt developers

License: EPL 1.0
"""

from __future__ import absolute_import, unicode_literals
import sys

from builtins import bytes  # from the future package
import numpy as np
try:
    import scipy
except ImportError:  # scipy is not installed
    SCIPY_INSTALLED = False
else:
    SCIPY_INSTALLED = True
    del scipy
    from scipy.optimize import approx_fprime
    try:
        from scipy.optimize import OptimizeResult
    except ImportError:
        # in scipy 0.14 Result was renamed to OptimzeResult
        from scipy.optimize import Result
        OptimizeResult = Result

import cyipopt


class IpoptProblemWrapper(object):
    """Class used to map an scipy minimize definition to a cyipopt problem.

    Parameters
    ==========
    fun : callable
        The objective function to be minimized: ``fun(x, *args, **kwargs) ->
        float``.
    args : tuple, optional
        Extra arguments passed to the objective function and its derivatives
        (``fun``, ``jac``, ``hess``).
    kwargs : dictionary, optional
        Extra keyword arguments passed to the objective function and its
        derivatives (``fun``, ``jac``, ``hess``).
    jac : callable, optional
        The Jacobian of the objective function: ``jac(x, *args, **kwargs) ->
        ndarray, shape(n, )``. If ``None``, SciPy's ``approx_fprime`` is used.
    hess : callable, optional
        If ``None``, the Hessian is computed using IPOPT's numerical methods.
        Explicitly defined Hessians are not yet supported for this class.
    hessp : callable, optional
        If ``None``, the Hessian is computed using IPOPT's numerical methods.
        Explicitly defined Hessians are not yet supported for this class.
    constraints : {Constraint, dict} or List of {Constraint, dict}, optional
        See https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.minimize.html
        for more information.
    eps : float, optional
        Epsilon used in finite differences.

    """
    def __init__(self, fun, args=(), kwargs=None, jac=None, hess=None,
                 hessp=None, constraints=(), eps=1e-8):
        if not SCIPY_INSTALLED:
            msg = 'Install SciPy to use the `IpoptProblemWrapper` class.'
            raise ImportError()
        self.fun_with_jac = None
        self.last_x = None
        if hess is not None or hessp is not None:
            msg = 'Using hessian matrixes is not yet implemented!'
            raise NotImplementedError(msg)
        if jac is None:
            jac = lambda x0, *args, **kwargs: approx_fprime(x0, fun, eps,
                                                            *args, **kwargs)
        elif jac is True:
            self.fun_with_jac = fun
        elif not callable(jac):
            raise NotImplementedError('jac has to be bool or a function')
        self.fun = fun
        self.jac = jac
        self.args = args
        self.kwargs = kwargs or {}
        self._constraint_funs = []
        self._constraint_jacs = []
        self._constraint_args = []
        self._constraint_kwargs = []
        if isinstance(constraints, dict):
            constraints = (constraints, )
        for con in constraints:
            con_fun = con['fun']
            con_jac = con.get('jac', None)
            con_args = con.get('args', [])
            con_kwargs = con.get('kwargs', [])
            if con_jac is None:
                con_jac = lambda x0, *args, **kwargs: approx_fprime(x0, con_fun, eps, *args, **kwargs)
            self._constraint_funs.append(con_fun)
            self._constraint_jacs.append(con_jac)
            self._constraint_args.append(con_args)
            self._constraint_kwargs.append(con_kwargs)
        # Set up evaluation counts
        self.nfev = 0
        self.njev = 0
        self.nit = 0

    def evaluate_fun_with_grad(self, x):
        if self.last_x is None or not np.all(self.last_x == x):
            self.last_x = x
            self.nfev += 1
            self.last_value = self.fun(x, *self.args, **self.kwargs)
        return self.last_value

    def objective(self, x):
        if self.fun_with_jac:
            return self.evaluate_fun_with_grad(x)[0]

        self.nfev += 1
        return self.fun(x, *self.args, **self.kwargs)

    def gradient(self, x, **kwargs):
        if self.fun_with_jac:
            return self.evaluate_fun_with_grad(x)[1]

        self.njev += 1
        return self.jac(x, *self.args, **self.kwargs)  # .T

    def constraints(self, x):
        con_values = []
        for fun, args in zip(self._constraint_funs, self._constraint_args):
            con_values.append(fun(x, *args))
        return np.hstack(con_values)

    def jacobian(self, x):
        con_values = []
        for fun, args in zip(self._constraint_jacs, self._constraint_args):
            con_values.append(fun(x, *args))
        return np.vstack(con_values)

    def intermediate(
            self,
            alg_mod,
            iter_count,
            obj_value,
            inf_pr,
            inf_du,
            mu,
            d_norm,
            regularization_size,
            alpha_du,
            alpha_pr,
            ls_trials
            ):

        self.nit = iter_count


def get_bounds(bounds):
    if bounds is None:
        return None, None
    else:
        lb = [b[0] for b in bounds]
        ub = [b[1] for b in bounds]
        return lb, ub


def get_constraint_bounds(constraints, x0, INF=1e19):
    if isinstance(constraints, dict):
        constraints = (constraints, )
    cl = []
    cu = []
    if isinstance(constraints, dict):
        constraints = (constraints, )
    for con in constraints:
        m = len(np.atleast_1d(con['fun'](x0, *con.get('args', []))))
        cl.extend(np.zeros(m))
        if con['type'] == 'eq':
            cu.extend(np.zeros(m))
        elif con['type'] == 'ineq':
            cu.extend(INF*np.ones(m))
        else:
            raise ValueError(con['type'])
    cl = np.array(cl)
    cu = np.array(cu)

    return cl, cu


def replace_option(options, oldname, newname):
    if oldname in options:
        if newname not in options:
            options[newname] = options.pop(oldname)


def convert_to_bytes(options):
    if sys.version_info >= (3, 0):
        for key in list(options.keys()):
            try:
                if bytes(key, 'utf-8') != key:
                    options[bytes(key, 'utf-8')] = options[key]
                    options.pop(key)
            except TypeError:
                pass


def minimize_ipopt(fun, x0, args=(), kwargs=None, method=None, jac=None,
                   hess=None, hessp=None, bounds=None, constraints=(),
                   tol=None, callback=None, options=None):
    """
    Minimize a function using ipopt. The call signature is exactly like for
    ``scipy.optimize.mimize``. In options, all options are directly passed to
    ipopt. Check [http://www.coin-or.org/Ipopt/documentation/node39.html] for
    details. The options ``disp`` and ``maxiter`` are automatically mapped to
    their ipopt-equivalents ``print_level`` and ``max_iter``.
    """
    if not SCIPY_INSTALLED:
        msg = 'Install SciPy to use the `minimize_ipopt` function.'
        raise ImportError(msg)

    _x0 = np.atleast_1d(x0)
    problem = IpoptProblemWrapper(fun, args=args, kwargs=kwargs, jac=jac,
                                  hess=hess, hessp=hessp,
                                  constraints=constraints)
    lb, ub = get_bounds(bounds)

    cl, cu = get_constraint_bounds(constraints, x0)

    if options is None:
        options = {}

    nlp = cyipopt.Problem(n=len(_x0),
                          m=len(cl),
                          problem_obj=problem,
                          lb=lb,
                          ub=ub,
                          cl=cl,
                          cu=cu)

    # python3 compatibility
    convert_to_bytes(options)

    # Rename some default scipy options
    replace_option(options, b'disp', b'print_level')
    replace_option(options, b'maxiter', b'max_iter')
    if b'print_level' not in options:
        options[b'print_level'] = 0
    if b'tol' not in options:
        options[b'tol'] = tol or 1e-8
    if b'mu_strategy' not in options:
        options[b'mu_strategy'] = b'adaptive'
    if b'hessian_approximation' not in options:
        if hess is None and hessp is None:
            options[b'hessian_approximation'] = b'limited-memory'
    for option, value in options.items():
        try:
            nlp.add_option(option, value)
        except TypeError as e:
            msg = 'Invalid option for IPOPT: {0}: {1} (Original message: "{2}")'
            raise TypeError(msg.format(option, value, e))

    x, info = nlp.solve(_x0)

    if np.asarray(x0).shape == ():
        x = x[0]

    return OptimizeResult(x=x,
                          success=info['status'] == 0,
                          status=info['status'],
                          message=info['status_msg'],
                          fun=info['obj_val'],
                          info=info,
                          nfev=problem.nfev,
                          njev=problem.njev,
                          nit=problem.nit)



# -*- coding: utf-8 -*-
"""
cyipot: Python wrapper for the Ipopt optimization package, written in Cython.

Copyright (C) 2012 Amit Aides
Author: Amit Aides <amitibo@tx.technion.ac.il>
URL: <http://http://code.google.com/p/cyipopt/>
License: EPL 1.0
"""
#
# Test the "ipopt" Python interface on the Hock & Schittkowski test problem
# #71. See: Willi Hock and Klaus Schittkowski. (1981) Test Examples for
# Nonlinear Programming Codes. Lecture Notes in Economics and Mathematical
# Systems Vol. 187, Springer-Verlag.
#
# Based on matlab code by Peter Carbonetto.
#

import numpy as np
import scipy.sparse as sps
import cyipopt


class HS071(object):

    def __init__(self):
        self.hs = sps.coo_matrix(np.tril(np.ones((4, 4))))

    def objective(self, x):
        return x[0]*x[3]*np.sum(x[0:3]) + x[2]

    def gradient(self, x):
        return np.array([
                    x[0]*x[3] + x[3]*np.sum(x[0:3]),
                    x[0]*x[3],
                    x[0]*x[3] + 1.0,
                    x[0]*np.sum(x[0:3])
                    ])

    def constraints(self, x):
        return np.array((np.prod(x), np.dot(x, x)))

    def jacobian(self, x):
        return np.concatenate((np.prod(x) / x, 2*x))

    def hessianstructure(self):
        return (self.hs.col, self.hs.row)

    def hessian(self, x, lagrange, obj_factor):
        H = obj_factor*np.array((
                (2*x[3], 0, 0, 0),
                (x[3],   0, 0, 0),
                (x[3],   0, 0, 0),
                (2*x[0]+x[1]+x[2], x[0], x[0], 0)))

        H += lagrange[0]*np.array((
                (0, 0, 0, 0),
                (x[2]*x[3], 0, 0, 0),
                (x[1]*x[3], x[0]*x[3], 0, 0),
                (x[1]*x[2], x[0]*x[2], x[0]*x[1], 0)))

        H += lagrange[1]*2*np.eye(4)

        return H[self.hs.row, self.hs.col]

    def intermediate(self, alg_mod, iter_count, obj_value, inf_pr, inf_du, mu,
                     d_norm, regularization_size, alpha_du, alpha_pr,
                     ls_trials):

        msg = "Objective value at iteration #{:d} is - {:g}"
        print(msg.format(iter_count, obj_value))


def main():
    x0 = [1.0, 5.0, 5.0, 1.0]

    lb = [1.0, 1.0, 1.0, 1.0]
    ub = [5.0, 5.0, 5.0, 5.0]

    cl = [25.0, 40.0]
    cu = [2.0e19, 40.0]

    nlp = cyipopt.Problem(n=len(x0), m=len(cl), problem_obj=HS071(), lb=lb,
                          ub=ub, cl=cl, cu=cu)

    nlp.addOption(b'mu_strategy', b'adaptive')
    nlp.addOption(b'tol', 1e-7)

    nlp.setProblemScaling(obj_scaling=2, x_scaling=[1, 1, 1, 1])
    nlp.addOption(b'nlp_scaling_method', b'user-scaling')

    x, info = nlp.solve(x0)

    msg = "Solution of the primal variables: x={:s}\n"
    print(msg.format(repr(x)))

    msg = "Solution of the dual variables: lambda={:s}\n"
    print(msg.format(repr(info['mult_g'])))

    msg = "Objective={:s}\n"
    print(msg.format(repr(info['obj_val'])))


if __name__ == '__main__':
    main()
